#ifndef MAIN_H
#define MAIN_H
#include <string>
#define GEN_PRINT_BUFSIZE 1024
extern int lang; // menufr.cc
extern int execution_in_progress ; // set to 1 during program execution
extern volatile bool kbd_interrupted; extern int esc_flag;
extern bool freeze;
int get_free_memory();
bool eqws(char * s,bool eval); // from main.cc, s must be at least 512 char
bool save_script(const char * filename,const ustl::string & s);
int run_script(const char* filename) ;
void erase_script();
void save(const char * fname); // save session
int restore_session(const char * fname);
void displaylogo();
ustl::string remove_path(const ustl::string & st);
void edit_script(const char * fname);
int find_color(const char * s); 
ustl::string khicas_state();
bool textedit(char * s);
void run(const char * s,int do_logo_graph_eqw=7);
bool stringtodouble(const ustl::string & s1,double & d);
void set_abort();
void clear_abort();
ustl::string remove_extension(const ustl::string &);

struct logo_turtle {
  double x,y;
  float theta; // theta is given in degrees or radians dep. on angle_mode
  int s;
  bool visible:1; // true if turtle visible
  bool mark:1; // true if moving marks
  bool direct; // true if rond/disque is done in the trigonometric direction
  char color;
  char turtle_length;
  int radius; // 0 nothing, >0 -> draw a plain disk 
  // bit 0-8=radius, bit9-17 angle1, bit 18-26 angle2, bit 27=1 filled  or 0 
  // <0 fill a polygon from previous turtle positions
  logo_turtle(): x(35),y(35),theta(0),visible(true),mark(true),direct(true),color(1),turtle_length(5),radius(0) {}
  inline bool equal_except_nomark(const logo_turtle &t) const {
    return x==t.x && y==t.y && turtle_length==t.turtle_length && s==t.s && radius==t.radius;
  }
};
inline int absint(int x){ return x<0?-x:x; }


// MICROPY & color compat
extern "C" {
  int do_confirm(const char * s);
  int ck_getkey(int * keyptr);
  int PRGM_GetKey(void);  
  void PRGM_GetKey_OS(unsigned char* p);
  int GetKeyWait_OS(int* column, int* row, int type_of_waiting, int timeout_period, int menu, unsigned short* keycode);
  void OS_InnerWait_ms(int ms);
  void HourGlass();  
  const char * read_file(const char * filename);
  int load_script(const char * filename,ustl::string & s);
  int do_file(const char *file);
  char * micropy_init(int stack_size,int heap_size);
  int micropy_eval(const char * line);
  void  mp_deinit();
  void mp_stack_ctrl_init();
  void process_freeze();
  extern int parser_errorline,parser_errorcol;
  void python_free();
  extern int pythonjs_stack_size,pythonjs_heap_size;
  extern char * python_heap,*pythonjs_static_heap;
  int python_init(int stack_size,int heap_size);

  void c_draw_rectangle(int x,int y,int w,int h,int c);
  void c_fill_rect(int x,int y,int w,int h,int c);
  void c_draw_line(int x0,int y0,int x1,int y1,int c);
  void c_draw_circle(int xc,int yc,int r,int color,bool q1,bool q2,bool q3,bool q4);
  void c_draw_filled_circle(int xc,int yc,int r,int color,bool left,bool right);
  void c_draw_polygon(int * x,int *y ,int n,int color);
  void c_draw_filled_polygon(int * x,int *y, int n,int xmin,int xmax,int ymin,int ymax,int color);
  void c_draw_arc(int xc,int yc,int rx,int ry,int color,double theta1, double theta2);
  void c_draw_filled_arc(int x,int y,int rx,int ry,int theta1_deg,int theta2_deg,int color,int xmin,int xmax,int ymin,int ymax,bool segment);
  void c_set_pixel(int x,int y,int c);
  int os_get_pixel(int x,int y);
  int c_draw_string(int x,int y,int c,int bg,const char * s,bool fake);
  int c_draw_string_small(int x,int y,int c,int bg,const char * s,bool fake);
  int c_draw_string_medium(int x,int y,int c,int bg,const char * s,bool fake);
  int select_item(const char ** ptr,const char * title,bool askfor1=true);
  // C conversion to gen from atomic data type 
  unsigned long long c_double2gen(double); 
  unsigned long long c_int2gen(int);
  // linalg on double matrices
  bool r_inv(double *,int n);
  bool r_rref(double *,int n,int m);
  double r_det(double *,int);
  struct double_pair {
    double r,i;
  } ;
  typedef struct double_pair c_complex;
  bool c_inv(c_complex *,int n);
  bool c_rref(c_complex *,int n,int m);
  c_complex c_det(c_complex *,int);
  bool c_egv(c_complex * x,int n); // eigenvectors
  bool c_eig(c_complex * x,c_complex * d,int n); // x eigenvect, d reduced mat
  bool c_proot(c_complex * x,int n); // poly root
  bool c_pcoeff(c_complex * x,int n); // root->coeffs
  bool c_fft(c_complex * x,int n,bool inverse); // FFT
  void turtle_freeze();
  void console_freeze();
  void console_output(const char * s,int l);
  const char * console_input(const char * msg1,const char * msg2,bool numeric,int ypos);
  void c_sprint_double(char * s,double d);
  void c_turtle_forward(double d);
  void c_turtle_left(double d);
  void c_turtle_up(int i);
  void c_turtle_goto(double x,double y);
  void c_turtle_cap(double x);
  int c_turtle_crayon(int i);
  void c_turtle_rond(int x,int y,int z);
  void c_turtle_disque(int x,int y,int z,int centered);
  void c_turtle_fill(int i);
  void c_turtle_fillcolor(double r,double g,double b,int entier);
  void c_turtle_getposition(double * x,double * y);
  void c_turtle_clear(int clrpos);
  void c_turtle_show(int visible);
  int c_turtle_getcap();
  void py_ck_ctrl_c();
  int micropy_ck_eval(const char *line);
  const char * caseval(const char * s);
  int kbd_convert(int r,int c);
  int kbd_filter(int key); // replace up/down/right/left/exe/exit with Nuumworks keycode
}

#endif
